import abc
import json
import time
from multiprocessing import Process, Manager
from threading import Thread
from typing import List, Optional, Dict

from .logger import Logger


class ParallelFn(abc.ABC):
    @abc.abstractmethod
    def apply(self, worker_name, log_fn, **kwargs) -> List:
        raise NotImplementedError()


class ParallelFnExecutor:
    def __init__(self, name, num_worker, use_thread=False, debug=False):
        self.name = name
        self.worker_count = num_worker
        self.__logger = Logger(f'{name}_executor') if debug else None
        self.__file_proxy: Optional[List] = None
        self.__use_thread = use_thread
        self.__fn_cls = None
        self.__debug = debug

    def log(self, message, error=False):
        if self.__logger is None:
            return
        if error:
            self.__logger.error(message)
        else:
            self.__logger.info(message)

    def log_fn(self, worker_name):
        def _log_fn(message, error=False):
            self.log(f'[{worker_name}]{message}', error)
        return _log_fn

    def __execute_job(self, worker_name, params):
        self.log(f"[{worker_name}] Started")

        start_ts = time.time()

        # execute fn
        data = self.__fn_cls().apply(worker_name, self.log_fn(worker_name), **params)

        # save result
        file_name = f"/tmp/{time.time()}_{worker_name}.json"
        with open(file_name, "w") as wd:
            wd.write(json.dumps(data, separators=(',', ':')))
        self.__file_proxy.append(file_name)

        end_ts = time.time()

        self.log(f"[{worker_name}] Done in {end_ts - start_ts:.2f}s")

    @property
    def __process_keys(self):
        for i in range(self.worker_count):
            yield f"{self.name}_exec_proc_{i}"

    def __split_jobs(self, static_params: Dict, dynamic_params: Dict):
        params = {**static_params}
        job_args = {}
        for k, v in dynamic_params.items():
            task_count = int(len(v) / self.worker_count)
            for i, proc in enumerate(self.__process_keys):
                values = v[i * task_count:] if i == self.worker_count - 1 else v[i * task_count:(i + 1) * task_count]
                if proc in job_args:
                    job_args[proc][k] = values
                else:
                    job_args[proc] = {k: values}
        for k, v in job_args.items():
            v.update(params)
        return job_args

    def __start_workers(self, params: Dict):
        worker_pool = []
        executor = Thread if self.__use_thread else Process
        for proc in self.__process_keys:
            proc = executor(target=self.__execute_job, args=(proc, params[proc]))
            proc.start()
            worker_pool.append(proc)
        for proc in worker_pool:
            proc.join()

    def __merge_result(self) -> List:
        result = []
        for file in self.__file_proxy:
            with open(file, "r") as rd:
                try:
                    result.extend(json.loads(rd.read()))
                except Exception as e:
                    self.log(f"Failed to read result {file}: {str(e)}", error=True)
        return result

    def execute(self, fn, static_params: Optional[Dict] = None, params: Optional[Dict] = None) -> List:
        self.__file_proxy = Manager().list()
        self.__fn_cls = fn
        self.__start_workers(self.__split_jobs(static_params, params))
        return self.__merge_result()



import logging

FILE_LOG_ENABLED = False
FILE_LOG_PATH = ''

DEBUG = logging.DEBUG


class Logger:
    def __init__(self, module_name, logging_level=logging.INFO):
        self.__logger = logging.getLogger('xcrawler')
        self.__logging_level = logging_level
        self.__module_name = module_name
        if len(self.__logger.handlers) == 0:
            if FILE_LOG_ENABLED:
                # File handler
                handler = logging.FileHandler(FILE_LOG_PATH)
                handler.setLevel(logging.DEBUG)
                handler.setFormatter(logging.Formatter(f'[%(asctime)s][%(levelname)s]%(message)s'))
                self.__logger.addHandler(handler)
            # Stream handler
            handler = logging.StreamHandler()
            handler.setLevel(logging.DEBUG)
            handler.setFormatter(logging.Formatter(f'%(message)s'))
            self.__logger.addHandler(handler)

            self.__logger.propagate = False
            self.__logger.setLevel(logging.DEBUG)

    # std flush
    def flush(self):
        pass

    # std log
    def write(self, buf):
        self.__logger.log(self.__logging_level, f'[{self.__module_name}] {buf.rstrip()}')

    def info(self, msg):
        self.__logger.info(f'[{self.__module_name}] {msg}')

    def error(self, msg):
        self.__logger.error(f'[{self.__module_name}] {msg}')



from itertools import islice, zip_longest
import csv


def read_csv_data_chunk(file_path, csv_headers, data_row=2, chunk_size=1000, encoding='utf-8'):
    """
        Lazy read csv file with chunk_size and select data base on column names
    """

    def jump_to_data_row_of_chunk():
        try:
            next(islice(reader, data_row - 2, None))  # jump to before data row
        except StopIteration:
            raise ValueError('Invalid data row input {}'.format(data_row))

    if data_row < 2:
        raise ValueError('No data found at row {}'.format(data_row))

    with open(file_path, encoding=encoding) as f:
        reader = csv.reader(f, csv.excel)
        jump_to_data_row_of_chunk()
        result = []
        for line in reader:
            row = dict(zip_longest(csv_headers, line))
            result.append(row)
            if len(result) == chunk_size:
                yield result
                result = []
        if len(result) > 0:  # in case size of last chunk is less than chunk size
            yield result


def read_csv_header(file_path, header_row=1, encoding='utf-8'):
    """
        Read the headers of csv file.
    """
    if header_row < 1:
        raise ValueError('Header row must be greater than or equals 1 , value={}'.format(header_row))
    with open(file_path, encoding=encoding) as f:
        reader = csv.reader(f, csv.excel)
        try:
            return next(islice(reader, header_row - 1, None))
        except StopIteration:
            raise ValueError('Can\'t find header row at line {}'.format(header_row))



from ..registry import Registry


class DatabaseConfig:
    host: str
    port: int
    user: str
    password: str
    database: str

    @property
    def postgre_connection_string(self):
        return f'postgresql://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}'


class RawDataDatabaseConfig(Registry, DatabaseConfig):
    def __init__(self):
        Registry.__init__(self, 'data_db')
        DatabaseConfig.__init__(self)


class MQConfig(Registry):
    host: str
    port: int
    user: str
    password: str

    def __init__(self):
        Registry.__init__(self, 'message_queue')



import json
import os
from contextlib import contextmanager
from typing import Dict
from typing import Optional, ContextManager

import requests

from .utils.logger import Logger


class RegistryConfig:
    def __init__(self, host: str, port: int):
        self.registry_host = host
        self.registry_port = port

    @staticmethod
    def from_env():
        host = os.environ.get('MP_REGISTRY_HOST', None)
        port = int(os.environ.get('MP_REGISTRY_PORT', -1))
        if not host or port < 0:
            raise ValueError("Please set registry host and port in environment variable")
        return RegistryConfig(host=host, port=port)


class Registry(object):
    def __init__(self, key, default_values: Optional[Dict] = None, config: Optional[RegistryConfig] = None):
        self.__data: Optional[Dict] = None
        if not key:
            raise ValueError("Need specify key of registry")
        self.__default_values = default_values
        self.__key = key
        self.__registry_config = config if config else RegistryConfig.from_env()
        self.__is_loaded = False
        self.__logger = Logger(f'{key}_registry')

    def __load_remote_configs(self):
        self.__is_loaded = False
        try:
            remote_config_url = f"http://{self.__registry_config.registry_host}:{self.__registry_config.registry_port}/get?key={self.__key}"
            response = requests.get(remote_config_url, timeout=3)
            self.__data = response.json()
            self.__is_loaded = True
        except Exception as e:
            self.__logger.error(f"Load remote configs failed: {str(e)}")
            self.__data = self.__default_values
        return self

    def __save_configs(self):
        try:
            endpoint = f"http://{self.__registry_config.registry_host}:{self.__registry_config.registry_port}/update"
            payload = {
                "key": self.__key,
                "value": json.dumps(self.__data)
            }
            response = requests.post(endpoint, json=payload)
            if response.status_code != 200:
                raise RuntimeError(f'Could not save registry values: [{response.status_code}] [{response.content.decode("utf-8")}]')
            data = response.json()
            if not data.get('success', False):
                raise RuntimeError(f'Could not save registry values: [{data}]')
        except Exception as e:
            self.__logger.error(f"Save remote configs failed: {str(e)}")
        return self

    def __getattribute__(self, item: str):
        if item.startswith('_Registry__'):
            return super(Registry, self).__getattribute__(item)
        if self.__data is None and not self.__is_loaded:
            self.__load_remote_configs()
        if item in self.__data:
            return self.__data[item]
        return super(Registry, self).__getattribute__(item)

    @contextmanager
    def __call__(self, *args, **kwargs) -> ContextManager["Registry"]:
        try:
            if self.__data is None and not self.__is_loaded:
                self.__load_remote_configs()
            yield self
        finally:
            self.__save_configs()

    def __setattr__(self, key, value):
        if key.startswith('_Registry__'):
            super(Registry, self).__setattr__(key, value)
        if self.__data and key in self.__data:
            self.__data[key] = value
        else:
            super(Registry, self).__setattr__(key, value)



from contextlib import contextmanager
from typing import TYPE_CHECKING, ContextManager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool

from .logger import Logger

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


class Database:
    __SESSION__ = None

    def __init__(self, connection_string: str, name='default', debug=False):
        self.connection_string = connection_string
        self.logger = Logger(f'{name}_db')
        self.debug = debug

    def init_db(self, base):
        # Create engine
        engine = create_engine(self.connection_string, echo=False, poolclass=NullPool)
        # Create meta
        base.metadata.create_all(engine)
        # Session maker
        self.__SESSION__ = sessionmaker(bind=engine)

    @contextmanager
    def __call__(self, *args, **kwargs) -> ContextManager["Session"]:
        session: "Session" = self.__SESSION__()
        try:
            yield session
            session.commit()
            if self.debug:
                self.logger.info("Committed to db")
        except Exception as e:
            session.rollback()
            self.logger.error(str(e))
            raise e
        finally:
            session.close()



from contextlib import contextmanager
from typing import Optional, Union, ContextManager

import pika
from pika import ConnectionParameters

from .logger import Logger


class _BasicPublisher:
    def __init__(self, logger, queue_name, channel):
        self.queue_name = queue_name
        self.channel = channel
        self.__logger = logger

    def enqueue(self, message: Union[bytes, str]):
        self.channel.basic_publish(
            exchange='',
            routing_key=self.queue_name,
            body=message if isinstance(message, bytes) else message.encode('utf-8'),
            properties=pika.BasicProperties(
                delivery_mode=2,  # make message persistent
            )
        )
        if self.__logger:
            self.__logger.info(f"Enqueued message, size={len(message) / 1024}KB")


class BasicPublish:
    def __init__(self, host: str, port: int, username: str, password: str, queue_name: str, debug: bool = False):
        self.connection: Optional[pika.BlockingConnection] = None
        self.channel = None
        self.conn_params = {
            "host": host,
            "port": port,
            "username": username,
            "password": password
        }
        self.queue_name = queue_name
        self.__is_connected = False
        self.__debug = debug
        self.__logger: Optional[Logger] = Logger(f"{queue_name}_queue") if debug else None

    @contextmanager
    def __call__(self, *args, **kwargs) -> ContextManager["_BasicPublisher"]:
        try:
            if not self.__is_connected:
                self.connect()
            yield _BasicPublisher(self.__logger, self.queue_name, self.channel)
        except Exception as e:
            if self.__debug:
                self.__logger.error(str(e))
            raise
        finally:
            self.disconnect()

    def connect(self):
        self.__setup_queue(ConnectionParameters(
            host=self.conn_params.get("host"),
            port=self.conn_params.get("port"),
            credentials=pika.PlainCredentials(
                username=self.conn_params.get("username"),
                password=self.conn_params.get("password")
            )
        ))
        self.__is_connected = True
        if self.__debug:
            self.__logger.info(f"Connected to RabbitMQ server")
        return self

    def disconnect(self):
        if not self.__is_connected:
            return
        self.__is_connected = False
        if self.connection:
            self.connection.close()
        if self.__debug:
            self.__logger.info(f"Disconnected from RabbitMQ server")

    def __setup_queue(self, params: ConnectionParameters):
        conn = pika.BlockingConnection(params)
        channel = conn.channel()
        channel.queue_declare(queue=self.queue_name, durable=True)
        # channel.exchange_declare(exchange=self.topic, exchange_type='topic')
        self.channel = channel
        self.connection = conn

    @property
    def is_connected(self):
        return self.__is_connected



import setuptools

setuptools.setup(
    name='micropattern',
    version="0.0.6",
    uthor='MicroPattern',
    description="A small helpers for our business",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
    install_requires=['pika==1.2.0', 'requests==2.24.0']
)
